<?php
/**
 * @file
 *   XLIFF import and export functionality
 */


/**
 * Import and export form for a specific node.
 *
 * @ingroup forms
 * @see xliff_form_submit()
 */
function xliff_form($form_state) {
  drupal_add_js(drupal_get_path('module', 'xliff') . '/xliff_admin.js');
  drupal_add_css(drupal_get_path('module', 'xliff') . '/xliff_admin.css');
  $form = array();
  $form['exportfs'] = array(
    '#type' => 'fieldset',
    '#title' => t('Export'),
    '#description' => t('All nodes of the selected content type will be included in XLIFF exports. Filtering options listed below can be used to filter the exported content.')
  );

  //Group 0: Content types
  $node_types     = node_get_types();
  $node_types_arr = array();

  foreach ($node_types AS $type) {
    if (translation_supported_type($type->type)) {
      $node_types_arr[$type->type] = $type->name;
    }
  }
  asort($node_types_arr); // Get sorted list of content type
  $form['exportfs']['content_type'] = array(
    '#type' => 'fieldset',
    '#title' => t('Content type')
  );
  $form['exportfs']['content_type']['node_type_options'] = array(
    '#type'          => 'select',
    '#title'         => t('Available Content types'),
    '#default_value' => variable_get('xliff_node_type', ''),
    '#options'       => $node_types_arr,
    '#required'      => TRUE,
    '#description'   => t('<b>Select node type for which you want to enable XLIFF export.</b><br/><em>Only those node types for which translation has been enabled will be displayed above and will be exported to XLIFF.</em>')
  );

  //Group 1: Type of export
  $form['exportfs']['export_type'] = array(
    '#type' => 'fieldset',
    '#title' => t('Export type')
  );
  $form['exportfs']['export_type']['export_type'] = array(
    '#type' => 'radios',
    '#default_value' => 0,
    '#options' => array(
      t('Export All'),
      t('Export "translate this" only'),
    ),
  );

  //Group 2: Filters
  $form['exportfs']['additional_filters'] = array(
    '#type' => 'fieldset',
    '#title' => t('Filters'),
    '#description' => t('These filters will apply to options selected in both of the above sections i.e. "Export Type" and "Outdated Translations".')
  );
  $form['exportfs']['additional_filters']['language_neutrals'] = array(
    '#type' => 'checkbox',
    '#title' => t('Include language neutral content. <strong>NOTE: Nodes that are language neutral will be updated to be English.</strong>'),
  );
  $form['exportfs']['additional_filters']['published_only'] = array(
    '#type' => 'checkbox',
    '#title' => t('Exclude un-published content.'),
  );
  $form['exportfs']['additional_filters']['export_using_tvid'] = array(
    '#type' => 'checkbox',
    '#title' => t('Only export nodes that are out of date based on TVID.'),
  );

  $term_form_element = _taxonomy_term_select(t('Filter by tags'), NULL, '', variable_get('xliff_vocab_option', NULL), NULL, TRUE, 'Ignore');
  if (count($term_form_element['#options']) > 1) {
    $form['exportfs']['additional_filters']['tagged_by_term'] = $term_form_element;
    $form['exportfs']['additional_filters']['tagged_by_term']['#weight'] = 1;  
  }
  else {
    $form['exportfs']['additional_filters']['xliff_no_terms'] =  array(
      '#value' => '<div class="form-item"><label>' . t('Grouping tag:') . '</label><div>' . t('Either there are no terms in the selected vocabulary or no vocabulary is selected in '.l('XLIFF settings', 'admin/content/xliff/settings').'.') . '</div></div>',
    );    
  }
  
  $languages    = language_list();
  $default_lang = variable_get('language_default', NULL);
  $lang_options = array('none' => "<None>");
  
  foreach ($languages AS $key => $value) {
    if ($key <> $default_lang->language) {
      $lang_options[$key] = $value->name;
    }
  }
  
  $form['exportfs']['export_target_language'] = array(
    '#type' => 'select',
    '#options' => $lang_options,
    '#default_value' => variable_get('xliff_target_language', ''),
    '#title' => t('Target language'),
    '#description' => t("You must specify a target language if you check '!tvid' above.",
                        array('!tvid' => $form['exportfs']['additional_filters']['export_using_tvid']['#title']))
  );
  $form['exportfs']['export'] = array(
    '#type' => 'submit',
    '#value' => t('Export as XLIFF'),
    '#description' => t('Export the title and body of the content for external translation.'),
    '#weight' => 2
  );

  $form['importfs'] = array(
    '#type' => 'fieldset',
    '#title' => t('Import')
  );
  $form['importfs']['xliff_file'] = array(
    '#type' => 'file',
    '#title' => t('XLIFF file'),
  );
  $form['importfs']['publish_type'] = array(
    '#type' => 'radios',
    '#title' => t('Publishing options'),
    '#default_value' => 2,
    '#options' => array(2 => t('Inherit'), 1 => t('Published'), 0 => t('Unpublished')),
  );
  $form['importfs']['relative_node_revisions'] = array(
    '#type' => 'checkbox',
    '#title' => t('Re-point node references to the translated versions of the referenced nodes (if translations exist).'),
  );
  
  $node_types     = node_get_types();
  $node_types_arr = array();

  foreach ($node_types AS $type) {
    if (translation_supported_type($type->type)) {
      $node_types_arr[$type->type] = $type->name;
    }
  }
  $form['importfs']['exclude_nr_types'] = array(
    '#type' => 'fieldset',
    '#title' => t('Exclude node-references for:'),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
  );
  $form['importfs']['exclude_nr_types']['excluded_nr_types'] = array(
    '#type' => 'checkboxes',
    '#title' => t('Node-references belonging to content-types selected below will not be repointed while updating translations'),
    '#default_value' => array(),
    '#options' => $node_types_arr,
  );
  $form['importfs']['import'] = array(
    '#type' => 'submit',
    '#value' => t('Import XLIFF'),
    '#description' => t('Replace the contents of the title and body of the content with the uploaded document.'),
  );

  $form['#attributes']['enctype'] = 'multipart/form-data';
  
  
  $form['noderefs'] = array(
    '#type' => 'fieldset',
    '#title' => t('Node References'),
    '#description' => t('Re-point node references to the translated versions of the referenced nodes (if translations exist). ' .
                        'This does the same as the Import checkbox above, but can be run at any time.')
  );
  $form['noderefs']['repoint'] = array(
    '#type' => 'submit',
    '#value' => t('Re-point Node References'),
  );

  return $form;
}

/**
 * Validation for XLIFF form.
 */
function xliff_form_validate($form, $form_state) {
  //If the user clicked the export button, validate the target language
  if ($form_state['clicked_button']['#value'] === $form['exportfs']['export']['#value']) {
    $tvid_text = $form['exportfs']['additional_filters']['export_using_tvid']['#title'];
    if ($form_state['values']['export_using_tvid'] == 1) {
      //User must choose a target language
      if ($form_state['values']['export_target_language'] == 'none') {
        form_set_error('export_target_language',
                       t("You must choose a target language if you check '!tvid'", array('!tvid' => $tvid_text)));
      }
    }
    else {
      //Don't allow target language to be set if "use TVID" was not checked
      if ($form_state['values']['export_target_language'] <> 'none') {
        form_set_error('export_target_language',
                       t("Target language is only applicable if you check '!tvid' " .
                         "Otherwise, choose '<None>'.", array('!tvid' => $tvid_text)));
      }
    }
  }
}


/**
 * Submit handler for XLIFF form.
 */
function xliff_form_submit($form, $form_state) {
  if ($form_state['values']['op'] == $form_state['values']['export']) {
    //Save variable so it can be used to populate form automatically
    variable_set('xliff_target_language', $form_state['values']['export_target_language']);
    
    if (!empty($form_state['values']['node_type_options'])) {
      variable_set('xliff_node_type', $form_state['values']['node_type_options']);
    }
    $output = xliff_export(
        $form_state['values']['export_type'],
        $form_state['values']['language_neutrals'],
        $form_state['values']['published_only'],
        $form_state['values']['tagged_by_term'],
        $form_state['values']['export_target_language'],
        $form_state['values']['export_using_tvid']
      );
    
    if ($output === FALSE) {
      drupal_set_message("All nodes are up-to-date. Nothing was exported.");
    }
    else {
      $content_type = variable_get('xliff_node_type', '');
      header('Content-type: application/xml; encoding=utf8');
      header('Content-Disposition: attachment; filename="' . $content_type . '.xliff"');
      echo $output;
      exit();
    }
  }
  elseif ($file = file_save_upload('xliff_file', array(), file_directory_path())) {
    $_SESSION['publish_type'] = $form_state['values']['publish_type'];
    $_SESSION['relative_node_revisions'] = $form_state['values']['relative_node_revisions'];
    $_SESSION['excluded_nr_types'] = $form_state['values']['excluded_nr_types'];
    xliff_import($file);
  }
  elseif ($form_state['values']['op'] == $form_state['values']['repoint']) {
    _repoint_node_references();
  }
}


/**
 * XLIFF export API for a nodes ready for translation.
 */
function xliff_export($export_type = 0, $language_neutrals = FALSE, $published_only = FALSE, $tagged_by_term = array(), $target_language = '', $use_tvid = FALSE) {
  $fin_arr = array();
  
  if (!variable_get('xliff_node_type', '')) {
    drupal_set_message(t("You have not selected a node type."), 'error');
    return;
  }

  $node_type = variable_get('xliff_node_type', '');
  $nids = array();
  $skipped_nids = array();
  
  //Tagged by term
  if (isset($tagged_by_term[''])) {
    unset($tagged_by_term['']);
  }
  
  //Basic export query: Export Type filedset
  $query = "SELECT n.nid FROM {node} n";

  //"Translate This"
  if ($export_type == 1) {
    $query .= " INNER JOIN {xliff_translate_this} xtt ON xtt.nid = n.nid AND xtt.field_translate_this = 1";
  }

  //Tagged by term
  if (count($tagged_by_term) > 0) {
    $query .= " INNER JOIN {term_node} tn ON tn.vid = n.vid AND tn.tid IN(" . implode(",", $tagged_by_term) . ")";
  }
  

  //WHERE conditions
  $query .= " WHERE n.type = '%s'" //restrict to selected content type
          . " AND (n.language = 'en' OR n.language = '')"; //nodes with default(en) or no language
    
  //Published only
  if ($published_only) {
    if (module_exists('xliff_language_suppress') && !empty($target_language)) {
      //If the source node is unpublished, export it anyway if
      //xliff_language_suppress has marked it with suppress=0
      $query .= " AND (n.status = 1 OR n.nid IN " .
                "(SELECT tnid FROM xliff_lang_suppress WHERE suppress <> 1 AND language='$target_language'))";
    }
    else {
      $query .= " AND n.status = 1 ";
    }
  }

  $result = db_query($query, $node_type);

  while ($data = db_fetch_object($result)) {
    $nids[] = $data->nid;
  }

  //Remove duplicate nids
  $nids = array_unique($nids);

  /**
   * Loop through resulting nodes and prepare XML format required for XLIFF conversion process
   */
  foreach ($nids as $nid) {
    if (module_exists('xliff_language_suppress')) {
      if (nid_marked_for_suppression($nid, $target_language)) {
        continue;
      }
    }

    $node = node_load($nid, NULL, TRUE);
    
    // Check for escaping language-neutral content
    if (!$language_neutrals && empty($node->language)) {
      continue;
    }
    
    $changed_fields = FALSE;
    $skip_node      = FALSE;
    
    //Only export this node if the translation in the target language
    //is out of date according to its TVID
    if ($use_tvid && $node->tnid > 0) {
      $translation_nid = db_result(db_query("SELECT nid FROM {node}
                                             WHERE tnid=%d AND language='%s'",
                                            $node->tnid, $target_language));

      if ($translation_nid) {
        $tvid = db_result(db_query("SELECT tvid FROM {xliff_tvid} WHERE nid=%d", $translation_nid));
        
        if ($tvid && ($tvid == $node->vid)) {
          //Source node hasn't changed since the last export. No need to export again.
          $skip_node = TRUE;
        }
        elseif ($tvid) {
          //Load the node revision that the existing translation was sourced from
          if ($translation_source = node_load($nid, $tvid, TRUE)) {
            //See if any text fields have changed in the source node
            $changed_fields = _get_modified_translatable_fields($node, $translation_source);
            if ($changed_fields === FALSE) {
              //Set the translation's tvid to the source node's current vid to
              //avoid running _translatable_fields_have_changed in the next export.
              db_query("UPDATE {xliff_tvid} SET tvid=%d WHERE nid=%d", $node->vid, $translation_nid);
              
              $skip_node = TRUE; //No text fields have changed, so we are not going to export it.
            }
          }
        }
      }
    }
    
    if (!$skip_node) {
      $node_id      = $node->nid . '-' . $node->vid;
      $dom_html     = "";
      $allowed_keys = _xliff_get_allowed_keys($node);
      
      //Loop through every field and include only textual fields in the XML
      foreach ($node as $key => $value) {
        if (in_array($key, $allowed_keys)) {
          if ($changed_fields && !in_array($key, $changed_fields)) {
            //Don't include this field in the XLIFF output.
            continue;
          }
          $tidy = new tidy;
          
          $html_value = '';
          if (is_array($value)) { // Also check if the field in of 'text' type only.          
            if (count($value) > 1) { //For multi-valued text fields(CCK)
              $html_value_arr = array();
              foreach ($value AS $v) {              
                $html_value_arr[] = '<mvt>' . html_entity_decode((string)$v['value'], ENT_QUOTES, 'utf-8') . '</mvt>';
              }
              $html_value = implode('', $html_value_arr);            
            }
            else { //Single valued text fields(CCK)
              $html_value = html_entity_decode((string)$value[0]['value'], ENT_QUOTES, 'utf-8');
            }
            $html_value = html_convert_bad_chars($html_value);
            $tidy->parseString($html_value, array("output-xhtml"   => TRUE,
                                                  "show-body-only" => TRUE,
                                                  "wrap"           => 0,
                                                  "new-blocklevel-tags" => "mvt"
                                                 ),
                               'utf8'
                              );
          }
          else { //Non-CCK fields: title, body, teaser.
            $html_value = (string)$value;
            $html_value = html_convert_bad_chars($html_value);
            $tidy->parseString($html_value, array("output-xhtml"   => TRUE,
                                                  "show-body-only" => TRUE,
                                                  "wrap"           => 0
                                                 )
                              );
          }
          //Clean & Repair HTML structure of field before appending it to XML
          $tidy->cleanRepair();
          
          //Append field's content to XML
          $dom_html .= "<$key>$tidy</$key>\n";
        }
      }

      //Convert named entities to numeric entities.
      $dom_html = html_convert_entities($dom_html);
      $dom_html = "<html node='$node_id'>$dom_html</html>";    
      $html     = new DOMDocument();
  
  
      /**
       * Load node's XML content
       */
      if (!$html->loadXML($dom_html)) {// fails  for special characters  < and > DO NOT IGNORE THIS CASE
        drupal_set_message(t("<strong>The HTML structure of node '!ntitle' (!nid) is malformed. Please look for errors. &lt; and &gt; characters are not allowed in node's content.</strong>",
                              array('!ntitle' => $node->title,
                                    '!nid'    => $node_id
                                    )
                             ),
                           'error'
                          );
        return;
      }
  
      // Load XSLT
      $xsl = new DOMDocument;
      $xsl->load(drupal_get_path('module', 'xliff') . '/xml2xliff.xsl');
      $proc = new XSLTProcessor();
      $proc->ImportStyleSheet($xsl);
      
      // Transform DOM w.r.t. XSLT
      $fin_arr[] = $proc->transformToXML($html);
  
      // If transformation unsuccessful, record it.
      if ($fin_arr[count($fin_arr) - 1] == FALSE) {
        $skipped_nids[] = $node_id . ': ' . l($node->title, 'node/' . $node->nid);
      }
    }
  }
  
  if (empty($fin_arr)) {
    return FALSE; //No nodes need to be exported
  }

  /**
   * Show nodes missing from XLIFF export.
   */
  if (count($skipped_nids) > 0) {
    $skipped_nids = '<ol><li>' . implode('</li><li>', $skipped_nids) . '</li></ol>';
    drupal_set_message('Following nodes/content could not be exported. Please check the content of these nodes and then try to export XLIFF again:' . $skipped_nids, 'error');
  }

  $fin_xliff = '';

  /**
   * Join multiple-nodes in $fin_arr into one homogenous content
   */  
  foreach ($fin_arr AS $fin) {
    $body = preg_split('/<xliff(.+)>/', (string)$fin);
    $body = preg_split('/<\/xliff>/',   (string)$body[1]);
    $fin_xliff .= $body[0];
  }
  $fin_xliff = preg_replace('/<file(.+)>(.+)<\/xliff>/si', '', (string)$fin_arr[0]) . $fin_xliff . '</xliff>';
  
  return $fin_xliff;
}


/**
 * Import an XLIFF file with one/multiple translation of existing node(s)
 * @param Path of the file to be imported
 */
function xliff_import($file = NULL) {
  
  /**
   * Load uploaded XLIFF
   */
  try {
    $dom_html = new SimpleXMLElement($file->filepath, 0, TRUE);
  }
  catch (Exception $e) {
    drupal_set_message(t("<strong>There was an error while loading file: <i>!filename</i></strong>",
                          array('!filename' => $file->filename)
                        ),
                       'error'
                      );
    return FALSE;
  }

  //XLIFF header grabbed
  $xliff_header =  preg_replace('/<file(.+)>(.+)<\/xliff>/si', '', (string)$dom_html->asXML());
  $node_count   = count($dom_html->file);

  
  /**
   * Add file upload record to DB
   */
  $result_insert = db_query(  "INSERT INTO {xliff_files} (`id` ,`file_path` ,`timestamp`) "
                            . "VALUES (NULL , '%s', %d)",
                              $file->filepath,
                              time()
                           );

  if ($result_insert) {
    $result_file = db_query("SELECT `id` FROM {xliff_files} WHERE `file_path` = '%s'", $file->filepath);

    if ($data_file = db_fetch_object($result_file)) {
      $mysql_values = '';
      
      /**
       * Insert target language content of each node in XLIFF file in DB
       */
      for ($i = 0; $i < $node_count; $i++) {
        //Detect 'target-language' attribute in current <file> element of imported XLIFF.
        $target_language = NULL;
        $fin_xliff       = NULL;

        foreach ($dom_html->file[$i]->attributes() as $a => $b) {
          if ($a == 'target-language') {
            $target_language = strtolower((string) $b);
            break;
          }
        }

        if ($target_language == NULL) {
          drupal_set_message(t("<strong>'target-language' attribute missing from <file> element!</strong>"),
                             'error'
                            );
          return FALSE;
        }

        // Fragment the actual loadable XML(XLIFF) within the <file> element.
        $fin_xliff     = $xliff_header . $dom_html->file[$i]->asXML() . '</xliff>';
        $mysql_values .= ", (NULL, '" . $data_file->id . "', 'en', '" .
                         db_escape_string($target_language) . "', '" .
                         db_escape_string($fin_xliff) . "', '1')";
      }

      //Insert query
      $mysql_values = trim($mysql_values, ',');
      $result_records_insert = db_query(  "INSERT INTO {xliff_import_queue} "
                                        . "(`id`, `fid`, `source_lang`, `target_lang`, `xliff_value`, `status`) "
                                        . "VALUES " . $mysql_values
                                       );
      file_delete($file->filepath);

      if ($result_records_insert) {
        drupal_goto('admin/content/xliff/process_import_queue/' . $data_file->id);
      }
      else {
        drupal_set_message(t('Inserting XLIFF records in DB failed'), 
                           'error'
                          );
        return FALSE;
        //DO CLEAN-UP of DB
      }
    }
    else {
      drupal_set_message(t('File id could not be fetched. Please try import again.'), 
                         'error'
                        );
      //error: file id could not be fetched
    }
  }
  else {
    drupal_set_message(t('File info could not be stored in the database'), 
                       'error'
                      );
    //error: file info could not be stored in database
  }
  return TRUE;
}

/**
 * XLIFF import process page
 * @param file id
 */
function process_xliff_import_queue_page($fid = NULL) {
  if ($fid && is_numeric($fid)) {
    $output = '<iframe src="' . base_path() . 'admin/content/xliff/process_queue/'
            . $fid . '" width="100%" height="300" style="border: 0 solid transparent;">'
            . '<b>'
            . t('Your browser does not support iframes')
            . '</b>'
            . '</iframe>';
    return $output;
  }
  else {
    drupal_set_message(t('Illegal value of file idenfifier.'), 
                       'error'
                      );
  }
}

/**
 * Import content of the file specified by file id in batches specified by batch_size
 * @param file id
 */
function process_xliff_import_queue($fid = NULL) {
  if ($fid && is_numeric($fid)) {

    //Fix default batch size
    $batch_size = 1;

    //Check for session variables from previous processing
    if (!isset($_SESSION['xliff_queue_count_' . $fid])) {
      $result = db_query("SELECT count(id) AS count FROM {xliff_import_queue} WHERE fid=%d AND status=1 GROUP BY fid", $fid);
      $data = db_fetch_object($result);
      $_SESSION['xliff_queue_count_' . $fid] = $data->count;
    }

    if ($_SESSION['xliff_queue_count_' . $fid] > 0) {
      $tmp_batch_size = ($_SESSION['xliff_queue_count_' . $fid]*5)/100;

      //Modify batch_size depending on no. of records
      if ($tmp_batch_size > 1) {
        $batch_size = ($tmp_batch_size < 10) ? $tmp_batch_size : 10;
      }
      else {
        $batch_size = 1;
      }

      $result = db_query("SELECT * FROM {xliff_import_queue} WHERE fid=%d AND status=1 LIMIT 0,%d", $fid, $batch_size);

      if ($result->num_rows) {
        // Display progress bar
        $result_status = db_query("SELECT count(id) AS count FROM {xliff_import_queue} WHERE fid=%d AND status=1", $fid);
        $data_status   = db_fetch_object($result_status);
        $percent       = 100 - ((int)((100 * $data_status->count) / $_SESSION['xliff_queue_count_' . $fid]));
        $output        = '<link type="text/css" rel="stylesheet" media="all" href="'
                      . base_path() . 'modules/system/system.css?F" />'
                      . '<div id="updateprogress" class="progress" style="">'
                      . '<div class="bar">'
                      . '<div class="filled" style="width: ' . $percent . '%;"></div>'
                      . '</div>'
                      . '<div class="percentage">' . $percent . '%</div>'
                      . '<div class="message">'
                      . t('Remaining !count of !total',
                           array('!count' => $data_status->count,
                                 '!total' => $_SESSION['xliff_queue_count_' . $fid]
                                )
                          )
                      . '.<br>&nbsp;</div></div>';
        echo $output;

        //Import content
        while ($data = db_fetch_object($result)) {
          $source_language = $data->source_lang;
          $target_language = $data->target_lang;

          // Load the fragmented XML(XLIFF)
          $xliff = new DOMDocument;
          $xliff->loadXML($data->xliff_value);

          // Load the XSLT to be used for transforming loaded XLIFF to parsable XML
          $xsl = new DOMDocument;
          $xsl->load(drupal_get_path('module', 'xliff') . '/xliff2xml.xsl');
          $proc = new XSLTProcessor();
          $proc->ImportStyleSheet($xsl);

          // Transform to XML Object
          $html = simplexml_import_dom($proc->transformToDoc($xliff));
          if ($html == NULL) {
              $_SESSION['xliff_error_log_' . $fid][] = "<strong>"
                                                     . t("DBid: !dataid. The structure of XLIFF <i>'!filename'</i> does not adhere to the expected XLIFF structure necessary for a successful import.",
                                                         array('!dataid'   => $data->id,
                                                               '!filename' => $file->filename
                                                              )
                                                        )
                                                     . "</strong>";
          }

          $curr_nid = 0;
          $curr_vid = 0;

          foreach ($html->attributes() as $a => $b) {
            if ($a == 'node') {
              $tmp = explode('-', $b);
              $curr_nid = $tmp[0];
              $curr_vid = $tmp[1];
              break;
            }
          }

          //Translated data will now be converted to Drupal node

          //The source node from which this translation was created
          $original_node = node_load($curr_nid, NULL, TRUE);

          //The node object for this translation
          $curr_node = NULL;

          if ($curr_nid > 0 && $curr_vid > 0 && $original_node) {
            $node_translations = translation_node_get_translations($curr_nid);
            $is_new = (array_key_exists($target_language, $node_translations)) ? FALSE : TRUE;
            // Set original node's language to source language of translation being imported
            $original_node->language = $source_language;
            
            // Create a new session 'create_revision_default' and use this session for default node revision creation.
            $_SESSION['create_revision_default'] = $original_node->nid;

            if ($is_new) {
              //Create a clone of the original node
              $curr_node = drupal_clone($original_node);
              
              //Set default values for the new node
              $curr_node->language               = $target_language;
              $curr_node->translation_source     = $original_node;
              $curr_node->comment_count          = 0; // New translated node: Set to default value
              $curr_node->moderate               = 0; // New translated node: Set to default value
              $curr_node->pathauto_perform_alias = 1; // Force an automatic alias to be created
              
              //Unset nid so that a new node will be created
              unset($curr_node->nid);
              unset($curr_node->vid);
              unset($curr_node->path);
            }
            else {
              //Load the existing translation
              $curr_node = node_load($node_translations[$target_language]->nid, NULL, TRUE);
              
              if (module_exists('i18n_auto_clone')) {
                if (i18n_auto_clone_is_clone($curr_node->tnid, $curr_node->language)) {
                  //If the existing node is an autoclone, create a new autoalias
                  $curr_node->pathauto_perform_alias = 1;
                }
              }

              //Create a new revision
              $curr_node->revision = TRUE;
            }

            /**
             * Set publishing status of translation being imported.
             * If status=2 then inherit from source node. Otherwise,
             * set as selected in importing options.
             */
            if ($_SESSION['publish_type'] != 2) {
              $curr_node->status = (int)$_SESSION['publish_type'];
            }
            
            //If xliff_language_suppress is enabled and this node is designated for
            //language suppression, override the status value set above.
            if (module_exists('xliff_language_suppress')) {
              if (nid_marked_for_suppression($curr_nid)) {
                //set publication status to whatever the "suppress" field is for the target language;
                if (nid_marked_for_suppression($curr_nid, $target_language)) {
                  $curr_node->status = 0;
                }
                else {
                  $curr_node->status = 1;
                }
              }
            }

            /**
             * Load node of target-language with translated content
             * Populate a new node with translated content
             */
            foreach ($html AS $key => $value) {
              if (array_key_exists($key, $curr_node)) {
                $tmp_val = $value->asXML();
                
                if (count($value) == 0) {
                  $tmp_val = $value;
                }
                
                $key_replace = "/<$key>|<\/$key>/si";
                
                if (is_array($curr_node->$key)) {
                  //Does not work for CCK fields which are created using flexifields module
                  $tmp_arr = $curr_node->$key;
                  $tmp_val = preg_replace($key_replace, '', (string)$tmp_val);

                  //Check if current field has multi-valued content
                  if (strstr($tmp_val, '<mvt>')) {
                    //Process multi-valued content
                    $tmp_val     = substr($tmp_val, 5, -6);
                    $tmp_val_arr = explode('</mvt><mvt>', $tmp_val);
                    $mvt_arr     = array();
                    
                    foreach ($tmp_val_arr as $tmp_val_key => $a) {
                      $mvt_arr[] = array('value' => $a,
                                         'format' => $tmp_arr[$tmp_val_key]['format']);
                    }
                    $curr_node->$key = $mvt_arr;
                  }
                  else {
                    //Process single-valued content
                    $tmp_arr[0]['value'] = $tmp_val;
                    $curr_node->$key     = $tmp_arr;
                  }
                  unset($tmp_arr, $mvt_arr, $tmp_val_arr);
                }
                else {
                  $curr_node->$key = preg_replace($key_replace, '', (string)$tmp_val);
                }
                unset($tmp_val);
              }
            }

            /**
             * Process node reference fields
             */
            if ($_SESSION['relative_node_revisions']) {
              $nr_result = db_query('SELECT cnfi.field_name FROM {content_node_field_instance} cnfi'
                           . ' INNER JOIN {content_node_field} cnf ON cnf.field_name = cnfi.field_name AND cnf.type="nodereference"'
                           . ' WHERE cnfi.type_name="%s"',
                           $curr_node->type
                                   );

              while ($nr_data = db_fetch_object($nr_result)) {
                $field = $nr_data->field_name;
                if (isset($curr_node->$field)) {
                  //While updating translation, skip processing this node-reference field
                  //if referenced content type is marked to skip node-references repointing.
                  $field_settings = content_fields($field, $curr_node->type);
                  $skip_repointing = FALSE;                    
                  if (!$is_new) {
                    foreach ($_SESSION['excluded_nr_types'] AS $key => $val) {
                      if (($val === $key) && ($field_settings['referenceable_types'][$key] === $val)) {
                        $skip_repointing = TRUE;
                        break;
                      }
                    }
                    if ($skip_repointing) {
                      continue;
                    }
                  }

                  //Use field data from node in english language to repoint node-references.
                  $curr_node->$field = $original_node->$field;

                  if (is_array($curr_node->$field)) {
                    foreach ($curr_node->$field AS $nr_key => $nr_nid) {
                      //process nodereferences
                      $nr_nid_result  = db_query('SELECT type FROM {node} WHERE nid=%d', $nr_nid['nid']);
                      $nr_nid_data    = db_fetch_object($nr_nid_result);
                      $nr_nid['type'] = $nr_nid_data->type;

                      if (translation_supported_type($nr_nid['type'])) {
                        $nr_nid_translations = translation_node_get_translations($nr_nid['nid']);

                        if (array_key_exists($curr_node->language, $nr_nid_translations)) {
                          $tmp_arr            = &$curr_node->$field;
                          $tmp_arr[$nr_key]['nid'] = $nr_nid_translations[$curr_node->language]->nid;
                        }
                      }
                    }
                  }
                }
              }
            }
            
            //Before saving the translation, set all synchronized fields to the
            //value in the source node so we don't sync old, bad data.
            foreach (i18nsync_node_fields($curr_node->type) as $sync_field) {
              $curr_node->{$sync_field} = $original_node->{$sync_field};
            }

            //Allow other modules to alter the imported node before saving
            drupal_alter('xliff_import_node', $curr_node);
            
            node_save($curr_node);
            
            //Set tvid for $curr_node to the vid of the source node in the XLIFF
            $exists = db_result(db_query("SELECT 1 FROM {xliff_tvid} WHERE nid=%d", $curr_node->nid));
            if ($exists) {
              db_query("UPDATE {xliff_tvid} SET tvid=%d WHERE nid=%d", $curr_vid, $curr_node->nid);
            }
            else {
              db_query("INSERT INTO {xliff_tvid} (nid, tvid) VALUES (%d, %d)", $curr_node->nid, $curr_vid);
            }
            
            unset($fin_xliff, $xliff, $xsl, $proc, $curr_node, $tmp_arr);
            //Delete record from database
            db_query("UPDATE {xliff_import_queue} SET `status` = '0' WHERE `id` = %d", $data->id);
          }
          else {
            $_SESSION['xliff_error_log_' . $fid][] = t('<strong>(node id (nid):%current_node_id, current version id (vid):%current_node_version_id) Invalid node reference encountered.</strong>',
                                                        array('%current_node_id' => $curr_nid,
                                                              '%current_node_version_id' => $curr_vid
                                                             )
                                                       );
            // Toggle the comments below to keep/unkeep log of translations with import error. 'status': 0 = delete, 2 = keep
            db_query("UPDATE {xliff_import_queue} SET `status` = '0' WHERE `id` = %d", $data->id);
            //db_query("UPDATE {xliff_import_queue} SET `status` = '2' WHERE `id` = %d", $data->id);
          }
        }
        $result_delete_row = db_query("DELETE FROM {xliff_import_queue} WHERE `status` = 0");
        echo '<script type="text/javascript">self.location.reload()</script>';
      }
      else {
        if (isset($_SESSION['xliff_queue_count_' . $fid])) {
          $result_delete_row = db_query("DELETE FROM {xliff_files} WHERE `id` = %d", $fid);
          unset($_SESSION['xliff_queue_count_' . $fid]);
          $output = '<link type="text/css" rel="stylesheet" media="all" href="' . base_path() . 'modules/system/system.css?F" />'
                  . '<div id="updateprogress" class="progress" style="">'
                  . '<div class="bar">'
                  . '<div class="filled" style="width: 100%;"></div>'
                  . '</div>'
                  . '<div class="percentage">100%</div>'
                  . '<div class="message">'
                  . t("Import process completed.")
                  . '</div>';
          if (count($_SESSION['xliff_error_log_' . $fid]) > 0) {
            $output .= '<div class="error-message">'
                     . t('There were some errors while importing translations. For details click !atag here',
                         array('!atag' => '<a onclick="window.top.location.href=\''
                                          . base_path()
                                          . 'admin/content/xliff/import_errors/'
                                          . $fid
                                          . '\'" href="javascript:void(0)">'
                                )
                        )
                      . '</a>.</div>';
          }
          else {
            $output .= '<div class="message"><a onclick="window.top.location.href=\''
                     . base_path()
                     . 'admin/content/xliff\'" href="javascript:void(0)" >'
                     . t('Import another file')
                     . '</a></div>';
          }
          $output .= '</div>';
          
          _repoint_node_references();
          
          echo $output;
        }
        else {
          $_SESSION['xliff_error_log_' . $fid][] = '<div class="message error">' . t('No records found for the file in reference') . '</div>';
          unset($_SESSION['xliff_queue_count_' . $fid]);
        }
      }
    }
    else {
      echo '<div class="message error">' . t('No queued translations were found in the process queue for the selected file.') . '</div>';
      unset($_SESSION['xliff_queue_count_' . $fid]);
    }
  }
  else {
    echo '<div class="message error">' . t('Invalid file id:') . $fid . '</div>';
    unset($_SESSION['xliff_queue_count_' . $fid]);
  }
}

/**
 * Update node references so that nodes reference nodes of the same language.
 * Most references will be re-pointed during the import process, but this
 * function is called at the end of the import to handle references that couldn't
 * be repointed because of timing/order.
 */
function _repoint_node_references() {
  $node_ref_fields = db_query("SELECT field_name FROM {content_node_field} WHERE type = 'nodereference'");
  $updated = 0;
  $failed  = 0;

  while ($node_ref_field = db_fetch_array($node_ref_fields)) {
    $field = content_fields($node_ref_field['field_name']);
    if (!$field) {
      continue; //This is an inactive field. Skip it.
    }
    
    $table_info   = content_database_info($field);
    $select_delta = $field['multiple'] > 0
                      ? ', ref_table.delta AS delta'
                      : '';
    
    //Query to get node references with mismatched languages
    $sql = "
      SELECT referencing_node.language AS lang, referencing_node.vid AS vid,
             referenced_node.tnid AS tnid" . $select_delta . "
      FROM {node} AS referencing_node
      INNER JOIN {{$table_info['table']}} AS ref_table ON referencing_node.vid = ref_table.vid
      INNER JOIN {node} AS referenced_node ON ref_table.{$table_info['columns']['nid']['column']} = referenced_node.nid
      WHERE referencing_node.language <> '' AND referenced_node.language <> ''
        AND referencing_node.language <> referenced_node.language
    ";
    $mismatched_refs = db_query($sql);
    
    while ($mismatched_ref = db_fetch_array($mismatched_refs))  {
      $translations = translation_node_get_translations($mismatched_ref['tnid']);
      if (isset($translations[$mismatched_ref['lang']])) {
        //Translation exists, so repoint the reference
        $translation = $translations[$mismatched_ref['lang']];
        $delta = isset($mismatched_ref['delta'])
                   ? " AND delta={$mismatched_ref['delta']}"
                   : '';

        $sql = "UPDATE {{$table_info['table']}}
                SET {$table_info['columns']['nid']['column']}={$translation->nid}
                WHERE vid={$mismatched_ref['vid']}" . $delta;

        $result = update_sql($sql);

        if (!$result['success']) {
          $failed++;
        }
        else {
          $updated++;
        }
      }
    }
  }
  if ($updated > 0) {
    drupal_set_message("Successfully re-pointed $updated node references.");
  }
  if ($failed > 0) {
    drupal_set_message("Tried and failed to re-point $failed node references.");
  }
  if ($updated == 0 && $failed == 0) {
    drupal_set_message("No node references needed to be repointed.");
  }
}

/**
 * Display errors occured during export/import process
 */
function xliff_import_errors_page($fid) {
  $output = '';
  if (isset($_SESSION['xliff_error_log_' . $fid])) {
    //If there are a lot of errors then provide a downloadable file instead of error-messages
    if (count($_SESSION['xliff_error_log_' . $fid]) <= 5) {
      $output .= '<div class="error message">' . implode('<br/>', $_SESSION['xliff_error_log_' . $fid]) . '</div>';
    }
    else {
      $error_log_data = 'Error: ' . implode("\r\n\r\nError: ", $_SESSION['xliff_error_log_' . $fid]);      
      $strongs = array("<strong>", "</strong>", "<em>", "</em>");
      $error_log_data = str_ireplace($strongs, "", $error_log_data);
      
      $filepath = file_create_path('xliff_error_log.txt');
      $filepath = base_path() . file_save_data($error_log_data, $filepath, FILE_EXISTS_REPLACE);
      
      $output .= t("Click") . " <a href='$filepath'>" . t("here") . "</a> " . t("to download the error log.");
    }
    unset($_SESSION['xliff_error_log_' . $fid]);
  }
  else {
    $output .= "<div class='message error'>" . t("No errors found") . "</div>";
  }

  $output .= '<div class="message"><a href="' . base_path()
           . 'admin/content/xliff">' . t('Import a file') . '</a></div>';

  return $output;
}


/**
 * Generate XLIFF node settings form.
 */
function xliff_settings_form(&$form_state) {
  $form = array(
    '#action' => url('admin/content/xliff/settings'),
  );
  
  $vocab_options = array();
  $vocabs = taxonomy_get_vocabularies();
  foreach ($vocabs as $vocab) {
    $vocab_options[$vocab->vid] = $vocab->name;
  }

  if (count($vocab_options) > 0) {
    $form['xliff_vocab_options'] =  array(
      '#type' => 'select',
      '#title' => t('Select a vocabulary to use for the "Filter By Tags" option in the XLIFF Export'),
      '#default_value' => variable_get('xliff_vocab_option', NULL),
      '#options' => $vocab_options,
    );
  }
  else {
    $form['xliff_no_vocab'] =  array(
      '#value' => '<div class="form-item"><label>' . t('Select a vocabulary:') . '</label><div>' . t('No vocabularies are configured. Please create a vocabulary first.') . '</div></div>',
    );
  }
  
  $form['submit'] = array(
    '#type'  => 'submit',
    '#value' => t('Save'),
  );
  return $form;
}


function xliff_settings_form_submit($form, &$form_state) {
  if (!empty($form_state['values']['xliff_vocab_options'])) {
    variable_set('xliff_vocab_option', $form_state['values']['xliff_vocab_options']);
  }
  drupal_set_message(t('XLIFF settings have been saved.'));
}



/* html_convert_bad_chars($string) -- convert special characters in hexadecimal form to
 * XML-compatible special characters.
 */ 
function html_convert_bad_chars($string) {
  //Converts bad unicode entries, that doesn't have a match in ascii to html entities.
  $string = mb_convert_encoding($string, "HTML-ENTITIES", "UTF-8");

  return preg_replace_callback('/&#([a-zA-Z][a-zA-Z0-9]+)/',
    'convert_bad_chars', $string);
}

/* Swap hexadecimal bad characters that crash loadXML() with its loadXML() friendly characters.   
 * If the bad character isn't in the lookup table, this function returns a blank, which
 * destroys the character in the output - this is probably the
 * desired behaviour when producing XML.
 */
function convert_bad_chars($matches) {
  
  static $bad_chr = array(
        "\x00" => "chr(0)",
        "\x01" => "chr(1)",
        "\x02" => "chr(2)",
        "\x03" => "chr(3)",
        "\x04" => "chr(4)",
        "\x05" => "chr(5)",
        "\x06" => "chr(6)",
        "\x07" => "chr(7)",
        "\x08" => "chr(8)",
        "\x09" => "chr(9)",
        "\x0a" => "chr(10)",
        "\x0b" => "chr(11)",
        "\x0c" => "chr(12)",
        "\x0d" => "chr(13)",
        "\x0e" => "chr(14)",
        "\x0f" => "chr(15)",
        "\x10" => "chr(16)",
        "\x11" => "chr(17)",
        "\x12" => "chr(18)",
        "\x13" => "chr(19)",
        "\x14" => "chr(20)",
        "\x15" => "chr(21)",
        "\x16" => "chr(22)",
        "\x17" => "chr(23)",
        "\x18" => "chr(24)",
        "\x19" => "chr(25)",
        "\x1a" => "chr(26)",
        "\x1b" => "chr(27)",
        "\x1c" => "chr(28)",
        "\x1d" => "chr(29)",
        "\x1e" => "chr(30)",
        "\x1f" => "chr(31)"
  );
  return isset($bad_chr[$matches[1]]) ? $bad_chr[$matches[1]] : '';
}

/* html_convert_entities($string) -- convert named HTML entities to
 * XML-compatible numeric entities.
 */
function html_convert_entities($string) {
  return preg_replace_callback('/&([a-zA-Z][a-zA-Z0-9]+);/',
    'convert_entity', $string);
}


/* Swap HTML named entity with its numeric equivalent. If the entity
 * isn't in the lookup table, this function returns a blank, which
 * destroys the character in the output - this is probably the
 * desired behaviour when producing XML. */
function convert_entity($matches) {
  static $table = array(
                  'quot'     => '&#34;',
                  'amp'      => '&#38;',
                  'lt'       => '&#60;',
                  'gt'       => '&#62;',
                  'OElig'    => '&#338;',
                  'oelig'    => '&#339;',
                  'Scaron'   => '&#352;',
                  'scaron'   => '&#353;',
                  'Yuml'     => '&#376;',
                  'circ'     => '&#710;',
                  'tilde'    => '&#732;',
                  'ensp'     => '&#8194;',
                  'emsp'     => '&#8195;',
                  'thinsp'   => '&#8201;',
                  'zwnj'     => '&#8204;',
                  'zwj'      => '&#8205;',
                  'lrm'      => '&#8206;',
                  'rlm'      => '&#8207;',
                  'ndash'    => '&#8211;',
                  'mdash'    => '&#8212;',
                  'lsquo'    => '&#8216;',
                  'rsquo'    => '&#8217;',
                  'sbquo'    => '&#8218;',
                  'ldquo'    => '&#8220;',
                  'rdquo'    => '&#8221;',
                  'bdquo'    => '&#8222;',
                  'dagger'   => '&#8224;',
                  'Dagger'   => '&#8225;',
                  'permil'   => '&#8240;',
                  'lsaquo'   => '&#8249;',
                  'rsaquo'   => '&#8250;',
                  'euro'     => '&#8364;',
                  'fnof'     => '&#402;',
                  'Alpha'    => '&#913;',
                  'Beta'     => '&#914;',
                  'Gamma'    => '&#915;',
                  'Delta'    => '&#916;',
                  'Epsilon'  => '&#917;',
                  'Zeta'     => '&#918;',
                  'Eta'      => '&#919;',
                  'Theta'    => '&#920;',
                  'Iota'     => '&#921;',
                  'Kappa'    => '&#922;',
                  'Lambda'   => '&#923;',
                  'Mu'       => '&#924;',
                  'Nu'       => '&#925;',
                  'Xi'       => '&#926;',
                  'Omicron'  => '&#927;',
                  'Pi'       => '&#928;',
                  'Rho'      => '&#929;',
                  'Sigma'    => '&#931;',
                  'Tau'      => '&#932;',
                  'Upsilon'  => '&#933;',
                  'Phi'      => '&#934;',
                  'Chi'      => '&#935;',
                  'Psi'      => '&#936;',
                  'Omega'    => '&#937;',
                  'alpha'    => '&#945;',
                  'beta'     => '&#946;',
                  'gamma'    => '&#947;',
                  'delta'    => '&#948;',
                  'epsilon'  => '&#949;',
                  'zeta'     => '&#950;',
                  'eta'      => '&#951;',
                  'theta'    => '&#952;',
                  'iota'     => '&#953;',
                  'kappa'    => '&#954;',
                  'lambda'   => '&#955;',
                  'mu'       => '&#956;',
                  'nu'       => '&#957;',
                  'xi'       => '&#958;',
                  'omicron'  => '&#959;',
                  'pi'       => '&#960;',
                  'rho'      => '&#961;',
                  'sigmaf'   => '&#962;',
                  'sigma'    => '&#963;',
                  'tau'      => '&#964;',
                  'upsilon'  => '&#965;',
                  'phi'      => '&#966;',
                  'chi'      => '&#967;',
                  'psi'      => '&#968;',
                  'omega'    => '&#969;',
                  'thetasym' => '&#977;',
                  'upsih'    => '&#978;',
                  'piv'      => '&#982;',
                  'bull'     => '&#8226;',
                  'hellip'   => '&#8230;',
                  'prime'    => '&#8242;',
                  'Prime'    => '&#8243;',
                  'oline'    => '&#8254;',
                  'frasl'    => '&#8260;',
                  'weierp'   => '&#8472;',
                  'image'    => '&#8465;',
                  'real'     => '&#8476;',
                  'trade'    => '&#8482;',
                  'alefsym'  => '&#8501;',
                  'larr'     => '&#8592;',
                  'uarr'     => '&#8593;',
                  'rarr'     => '&#8594;',
                  'darr'     => '&#8595;',
                  'harr'     => '&#8596;',
                  'crarr'    => '&#8629;',
                  'lArr'     => '&#8656;',
                  'uArr'     => '&#8657;',
                  'rArr'     => '&#8658;',
                  'dArr'     => '&#8659;',
                  'hArr'     => '&#8660;',
                  'forall'   => '&#8704;',
                  'part'     => '&#8706;',
                  'exist'    => '&#8707;',
                  'empty'    => '&#8709;',
                  'nabla'    => '&#8711;',
                  'isin'     => '&#8712;',
                  'notin'    => '&#8713;',
                  'ni'       => '&#8715;',
                  'prod'     => '&#8719;',
                  'sum'      => '&#8721;',
                  'minus'    => '&#8722;',
                  'lowast'   => '&#8727;',
                  'radic'    => '&#8730;',
                  'prop'     => '&#8733;',
                  'infin'    => '&#8734;',
                  'ang'      => '&#8736;',
                  'and'      => '&#8743;',
                  'or'       => '&#8744;',
                  'cap'      => '&#8745;',
                  'cup'      => '&#8746;',
                  'int'      => '&#8747;',
                  'there4'   => '&#8756;',
                  'sim'      => '&#8764;',
                  'cong'     => '&#8773;',
                  'asymp'    => '&#8776;',
                  'ne'       => '&#8800;',
                  'equiv'    => '&#8801;',
                  'le'       => '&#8804;',
                  'ge'       => '&#8805;',
                  'sub'      => '&#8834;',
                  'sup'      => '&#8835;',
                  'nsub'     => '&#8836;',
                  'sube'     => '&#8838;',
                  'supe'     => '&#8839;',
                  'oplus'    => '&#8853;',
                  'otimes'   => '&#8855;',
                  'perp'     => '&#8869;',
                  'sdot'     => '&#8901;',
                  'lceil'    => '&#8968;',
                  'rceil'    => '&#8969;',
                  'lfloor'   => '&#8970;',
                  'rfloor'   => '&#8971;',
                  'lang'     => '&#9001;',
                  'rang'     => '&#9002;',
                  'loz'      => '&#9674;',
                  'spades'   => '&#9824;',
                  'clubs'    => '&#9827;',
                  'hearts'   => '&#9829;',
                  'diams'    => '&#9830;',
                  'nbsp'     => '&#160;',
                  'iexcl'    => '&#161;',
                  'cent'     => '&#162;',
                  'pound'    => '&#163;',
                  'curren'   => '&#164;',
                  'yen'      => '&#165;',
                  'brvbar'   => '&#166;',
                  'sect'     => '&#167;',
                  'uml'      => '&#168;',
                  'copy'     => '&#169;',
                  'ordf'     => '&#170;',
                  'laquo'    => '&#171;',
                  'not'      => '&#172;',
                  'shy'      => '&#173;',
                  'reg'      => '&#174;',
                  'macr'     => '&#175;',
                  'deg'      => '&#176;',
                  'plusmn'   => '&#177;',
                  'sup2'     => '&#178;',
                  'sup3'     => '&#179;',
                  'acute'    => '&#180;',
                  'micro'    => '&#181;',
                  'para'     => '&#182;',
                  'middot'   => '&#183;',
                  'cedil'    => '&#184;',
                  'sup1'     => '&#185;',
                  'ordm'     => '&#186;',
                  'raquo'    => '&#187;',
                  'frac14'   => '&#188;',
                  'frac12'   => '&#189;',
                  'frac34'   => '&#190;',
                  'iquest'   => '&#191;',
                  'Agrave'   => '&#192;',
                  'Aacute'   => '&#193;',
                  'Acirc'    => '&#194;',
                  'Atilde'   => '&#195;',
                  'Auml'     => '&#196;',
                  'Aring'    => '&#197;',
                  'AElig'    => '&#198;',
                  'Ccedil'   => '&#199;',
                  'Egrave'   => '&#200;',
                  'Eacute'   => '&#201;',
                  'Ecirc'    => '&#202;',
                  'Euml'     => '&#203;',
                  'Igrave'   => '&#204;',
                  'Iacute'   => '&#205;',
                  'Icirc'    => '&#206;',
                  'Iuml'     => '&#207;',
                  'ETH'      => '&#208;',
                  'Ntilde'   => '&#209;',
                  'Ograve'   => '&#210;',
                  'Oacute'   => '&#211;',
                  'Ocirc'    => '&#212;',
                  'Otilde'   => '&#213;',
                  'Ouml'     => '&#214;',
                  'times'    => '&#215;',
                  'Oslash'   => '&#216;',
                  'Ugrave'   => '&#217;',
                  'Uacute'   => '&#218;',
                  'Ucirc'    => '&#219;',
                  'Uuml'     => '&#220;',
                  'Yacute'   => '&#221;',
                  'THORN'    => '&#222;',
                  'szlig'    => '&#223;',
                  'agrave'   => '&#224;',
                  'aacute'   => '&#225;',
                  'acirc'    => '&#226;',
                  'atilde'   => '&#227;',
                  'auml'     => '&#228;',
                  'aring'    => '&#229;',
                  'aelig'    => '&#230;',
                  'ccedil'   => '&#231;',
                  'egrave'   => '&#232;',
                  'eacute'   => '&#233;',
                  'ecirc'    => '&#234;',
                  'euml'     => '&#235;',
                  'igrave'   => '&#236;',
                  'iacute'   => '&#237;',
                  'icirc'    => '&#238;',
                  'iuml'     => '&#239;',
                  'eth'      => '&#240;',
                  'ntilde'   => '&#241;',
                  'ograve'   => '&#242;',
                  'oacute'   => '&#243;',
                  'ocirc'    => '&#244;',
                  'otilde'   => '&#245;',
                  'ouml'     => '&#246;',
                  'divide'   => '&#247;',
                  'oslash'   => '&#248;',
                  'ugrave'   => '&#249;',
                  'uacute'   => '&#250;',
                  'ucirc'    => '&#251;',
                  'uuml'     => '&#252;',
                  'yacute'   => '&#253;',
                  'thorn'    => '&#254;',
                  'yuml'     => '&#255;'
                       );
  // Entity not found? Destroy it.
  return isset($table[$matches[1]]) ? $table[$matches[1]] : '';
}



/**
 * Import and export form for a specific node.
 *
 * @ingroup forms
 * @see xliff_form_submit()
 */
function xliff_multilingual_export_form($form_state) {
  drupal_add_css(drupal_get_path('module', 'xliff') . '/xliff_admin.css');
  
  $form = array();
  $languages = language_list();
  $lang_options = array();
  foreach ($languages AS $key => $val) {
    $lang_options[$key] = $val->name;
  }

  //Language of export
  $form['language'] = array(
    '#type' => 'select',
    '#options' => $lang_options,
    '#title' => t('Select a language in which you want to export the XLIFF'),
  );
  
  //Group 0: Content types
  $node_types     = node_get_types();
  $node_types_arr = array();

  foreach ($node_types AS $type) {
    if (translation_supported_type($type->type)) {
      $node_types_arr[$type->type] = $type->name;
    }
  }
  $form['exportfs']['content_type'] = array(
    '#type' => 'fieldset',
    '#title' => t('Content type')
  );
  $form['exportfs']['content_type']['node_type_options'] = array(
    '#type'          => 'select',
    '#title'         => t('Available Content types'),
    '#default_value' => variable_get('xliff_node_type', ''),
    '#options'       => $node_types_arr,
    '#required'      => TRUE,
    '#description'   => t('<b>Select node type for which you want to enable XLIFF export.</b><br/><em>Only those node types for which translation has been enabled will be displayed above and will be exported to XLIFF.</em>')
  );

  $form['exportfs']['export'] = array(
    '#type' => 'submit',
    '#value' => t('Export as XLIFF')
  );

  return $form;
}


/**
 * Submit handler for XLIFF form.
 */
function xliff_multilingual_export_form_submit($form, $form_state) {
    
  if (!empty($form_state['values']['node_type_options'])) {
    variable_set('xliff_node_type', $form_state['values']['node_type_options']);
  }
  if ($output = xliff_multilingual_export($form_state['values']['language'])) {
    $content_type = variable_get('xliff_node_type', '');
    header('Content-type: application/xml; encoding=utf8');
    header('Content-Disposition: attachment; filename="' . $content_type . '.xliff"');
    echo $output;
    exit();
  }
  else {
    drupal_set_message(t('Multilingual export failed'),
                       'error'
                      );
  }
}


/**
 * XLIFF export API for a nodes ready for translation.
 */
function xliff_multilingual_export($language = NULL) {
  if ($language != NULL) {
    $fin_arr = array();

    if (!variable_get('xliff_node_type', '')) {
      drupal_set_message(t("<strong>You have not selected any node types!</strong>"),
                         'error'
                        );
      return;
    }

    $node_type = variable_get('xliff_node_type', '');
    $nids = array();
    
    $result = db_query(  "SELECT n.nid "
                       . "FROM {node} n "
                       . "WHERE n.language = '" . $language . "'"   
                       . "AND n.type = '%s'"
                       , $node_type
                      );

    while ($data = db_fetch_object($result)) {
      $nids[] = $data->nid;
    }

    $nids = array_unique($nids);

    foreach ($nids as $nid) {
      $node = node_load($nid);
    
      $node_id  = $node->nid . '-' . $node->vid;
      
      //DOM generation
      $dom_html = "";

      $allowed_keys = _xliff_get_allowed_keys($node);
      
      /**
       * Loop through every field and include only textual fields in the XML
       */
      foreach ($node as $key => $value) {
        if (in_array($key, $allowed_keys)) {
          $tidy = new tidy;
          $html_value = '';
          if (is_array($value)) { // Also check if the field in of 'text' type only.          
            if (count($value) > 1) { //For multi-valued text fields(CCK)
              $html_value_arr = array();
              foreach ($value AS $v) {              
                $html_value_arr[] = '<mvt>' . html_entity_decode((string)$v['value'], ENT_QUOTES, 'utf-8') . '</mvt>';
              }
              $html_value = implode('', $html_value_arr);            
            }
            else { //Single valued text fields(CCK)
              $html_value = html_entity_decode((string)$value[0]['value'], ENT_QUOTES, 'utf-8');
            }
            $html_value = html_convert_bad_chars($html_value);
            $tidy->parseString($html_value, array("output-xhtml"   => TRUE,
                "show-body-only" => TRUE,
                "wrap"           => 0,
                "new-blocklevel-tags" => "mvt"
                                                 ),
              'utf8'
                              );
          }
          else { //Non-CCK fields: title, body, teaser.
            $html_value = (string)$value;
            $html_value = html_convert_bad_chars($html_value);
            $tidy->parseString($html_value, array("output-xhtml"   => TRUE,
                "show-body-only" => TRUE,
                "wrap"           => 0
                                                 )
                              );
          }
          //Clean & Repair HTML structure of field before appending it to XML
          $tidy->cleanRepair();
          //Append field's content to XML
          $dom_html .= "<$key>$tidy</$key>\n";
        }
      }

      $dom_html = html_convert_entities($dom_html);
      $dom_html = "<html node='$node_id'>$dom_html</html>";
      $html     = new DOMDocument();

      if (!$html->loadXML($dom_html)) {// fails  for special characters  < and > DO NOT IGNORE THIS CASE
        drupal_set_message(t("<strong>The HTML structure of node '!ntitle' (!nid) is malformed. Please look for errors. &lt; and &gt; characters are not allowed in node's content.</strong>",
                             array('!ntitle' => $node->title,
                                   '!nid'    => $node_id
                                  )
                            ),
                           'error'
                          );
        return;
      }

      // Load XSLT
      $xsl = new DOMDocument;
      $xsl->load(drupal_get_path('module', 'xliff') . '/xml2xliff.xsl');
      $proc = new XSLTProcessor();
      $proc->ImportStyleSheet($xsl);

      // Transform DOM w.r.t. XSLT
      $fin_arr[] = $proc->transformToXML($html);
    }

    $fin_xliff = '';

    //change source language string in final XLIFF element
    $lang_str = 'source-language="' . $language . '"';
    
    foreach ($fin_arr AS $fin) {
      $body = preg_split('/<xliff(.+)>/', (string)$fin);
      $body = preg_split('/<\/xliff>/',   (string)$body[1]);
      $body[0] = str_replace('source-language="en"', $lang_str, $body[0]);
      $fin_xliff .= $body[0];
    }

    $fin_xliff = preg_replace('/<file(.+)>(.+)<\/xliff>/si', '', (string)$fin_arr[0]) . $fin_xliff . '</xliff>';
    return $fin_xliff;
  }
  else {
    drupal_set_message(t('Invalid language'),
                       'error'
                      );
    return FALSE;
  }
}


/**
 * Menu item callback. Displays a list of all nodes that exist in the
 * chosen menu, in XML format.
 */
function xliff_menu_item_nids() {
  $form = array();
  
  $form['description'] = array(
    '#value' => 'Outputs the nid and title of all nodes that exist in the chosen menu, in XML format.',
  );
  $options = menu_parent_options(menu_get_menus(), $item);
  $default = isset($_GET['menu']) && isset($_GET['mlid'])
               ? $_GET['menu'] . ':' . $_GET['mlid']
               : 'primary-links:0';

  $form['menu_root'] = array(
    '#type' => 'select',
    '#title' => t('Parent item'),
    '#default_value' => $default,
    '#options' => $options,
    '#attributes' => array('class' => 'menu-title-select'),
  );
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Submit'),
  );
  
  //If a menu has been selected, display the output
  if (isset($_GET['menu'])) {
    //Use the menu API to get all the nodes in the selected menu
    $menu_tree = menu_tree_all_data($_GET['menu']);
    if (!empty($_GET['mlid'])) {
      //If a subtree was selected, extract it from the full tree
      $menu_tree = xliff_menu_get_subtree($menu_tree, $_GET['mlid']);
    }

    menu_tree_collect_node_links($menu_tree, $node_links);
  
    //Generate XML describing the nodes
    $output = '<nodes>' . PHP_EOL;
    
    foreach ($node_links as $nid => $value) {
      $menu_item_data = array_shift($value);
      $output .= '  <node>' . PHP_EOL;
      
      $output .= '    <title>' . $menu_item_data['title'] . '</title>' . PHP_EOL;
      $output .= "    <nid>$nid</nid>" . PHP_EOL;
      
      $output .= '  </node>' . PHP_EOL;
    }
    
    $output .= '</nodes>';
    
    $form['output'] = array(
      '#type' => 'textarea',
      '#rows' => 50,
      '#title' => t('Nodes in @menu menu', array('@menu' => $_GET['menu'])),
      '#default_value' => $output,
    );
  }

  return $form;
}

function xliff_menu_item_nids_submit($form, &$form_state) {
  if (isset($form_state['values']['menu_root'])) {
    //menu_root is in the format 'menu-name:mlid'
    $menu_root = explode(':', $form_state['values']['menu_root']);
    $query = array('menu' => $menu_root[0], 'mlid' => $menu_root[1]);

    //Reload the current page with the menu and mlid query parameters
    drupal_goto($_GET['q'], $query);
  }
}

/**
 * Extract a specific subtree from a menu tree based on a menu link id (mlid)
 *
 * @param array $tree
 *   A menu tree data structure as returned by menu_tree_all_data() or menu_tree_page_data()
 * @param int $mlid
 *   The menu link id of the menu entry for which to return the subtree
 * @return array
 *   The found subtree, or NULL if no entry matched the mlid
 */
function xliff_menu_get_subtree($tree, $mlid) {
  // Check all top level entries
  foreach ($tree as $key => $element) {
    // Is this the entry we are looking for?
    if ($mlid == $element['link']['mlid'])  {
      // Yes, return while keeping the key
      return array($key => $element);
    }
    else {
      // No, recurse to children, if any
      if ($element['below']) {
        $submatch = xliff_menu_get_subtree($element['below'], $mlid);
        // Found wanted entry within the children?
        if ($submatch) {
          // Yes, return it and stop looking any further
          return $submatch;
        }
      }
    }
  }
  // No match at all
  return NULL;
}

/**
 * For the given node object, returns an array containing the textual fields
 * that are translatable via xliff.
 */
function _xliff_get_allowed_keys($node) {
  $keys = array();

  foreach ($node AS $key => $value) {
    //Look for CCK fields
    if (substr((string)$key, 0, 6) == 'field_') {
      $keys[] = $key;
    }
  }
  
  $allowed_keys = array('title', 'body', 'teaser');
  
  //Get textual field names other than title, body and teaser(CCK)
  if (db_table_exists('content_node_field_instance') && !empty($keys)) {
    $result_keys = db_query("SELECT field_name
                             FROM {content_node_field_instance}
                             WHERE field_name IN (" . db_placeholders($keys, 'varchar') . ")
                               AND widget_type IN ('text_textfield', 'text_textarea')
                               AND type_name = '{$node->type}'",
                            $keys);
  
    while ($data_keys = db_fetch_object($result_keys)) {
      $allowed_keys[] = $data_keys->field_name;
    }
  }
  
  return $allowed_keys;
}

/**
 * Compares the textual fields in the two given node objects.
 * Returns FALSE if no fields have changed.
 * Otherwise, returns an array of the changed fields.
 */
function _get_modified_translatable_fields($node1, $node2) {
  $modified_fields = array();
  
  //$node1 and $node2 should be two revisions of the same node
  if ($node1->nid <> $node2->nid) {
    //This isn't the most graceful error handling, but we don't want to proceed
    //if this function receives two different nodes to compare, because it will
    //cause data in the xliff_tvid table to be corrupted.
    print "_get_modified_translatable_fields() received invalid arguments: nid1={$node1->nid}; nid2={$node2->nid}";
    exit;
  }
  
  $allowed_keys = _xliff_get_allowed_keys($node1);
  
  foreach ($node1 as $key => $value) {
    if (in_array($key, $allowed_keys)) {
      if (is_array($value)) {
        foreach ($value as $k => $v) {              
          if (strcmp($node1->{$key}[$k]['value'], $node2->{$key}[$k]['value']) <> 0) {
            $modified_fields[] = $key;
          }
        }
      }
      else {
        //Non-CCK fields: title, body, teaser.
        if (strcmp($node1->{$key}, $node2->{$key}) <> 0) {
          $modified_fields[] = $key;
        }
      }
    }
  }
  
  return empty($modified_fields) ? FALSE : $modified_fields;
}